/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.framework.driver.mbus;

import java.io.IOException;
import java.text.ParseException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeoutException;

import org.openmuc.framework.config.ArgumentSyntaxException;
import org.openmuc.framework.config.ChannelScanInformation;
import org.openmuc.framework.config.DeviceScanInformation;
import org.openmuc.framework.data.DoubleValue;
import org.openmuc.framework.data.Flag;
import org.openmuc.framework.data.Record;
import org.openmuc.framework.driver.spi.ChannelRecordContainer;
import org.openmuc.framework.driver.spi.ChannelValueContainer;
import org.openmuc.framework.driver.spi.ConnectionException;
import org.openmuc.framework.driver.spi.DeviceConnection;
import org.openmuc.framework.driver.spi.DriverService;
import org.openmuc.framework.driver.spi.RecordsReceivedListener;
import org.openmuc.jmbus.MBusSap;
import org.openmuc.jmbus.VariableDataBlock;
import org.openmuc.jmbus.VariableDataResponse;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Class implementing the DriverService.<br>
 * This class maps the jmbus library to the Driver Interface of OpenMUC
 * 
 */
public class MBusDriver implements DriverService {
	private final static Logger logger = LoggerFactory.getLogger(MBusDriver.class);

	private final Map<String, ConnectionHandle> connections = new HashMap<String, ConnectionHandle>();

	@Override
	public String getId() {
		return "mbus";
	}

	@Override
	public String getDescription() {
		return "M-Bus (wired) is a protocol to read out meters.";
	}

	@Override
	public List<DeviceScanInformation> scanForDevices(String interfaceAddress, String scanFilter, String settings,
			int timeout) throws UnsupportedOperationException, ArgumentSyntaxException, IOException {
		throw new UnsupportedOperationException("Device scan not available for MBus.");
	}

	@Override
	public List<ChannelScanInformation> scanForChannels(DeviceConnection connection, int timeout)
			throws UnsupportedOperationException, ConnectionException {
		throw new UnsupportedOperationException("Channel scan not available for MBus.");
	}

	@Override
	public Object connect(String interfaceAddress, String deviceAddress, String settings, int timeout)
			throws ArgumentSyntaxException, ConnectionException {

		ConnectionHandle connectionHandle = connections.get(interfaceAddress);

		if (connectionHandle == null) {

			MBusSap mBusSap;
			try {
				mBusSap = new MBusSap(interfaceAddress, settings, 0);
			} catch (IOException e1) {
				throw new ConnectionException("Unable to bind local interface: " + interfaceAddress);
			} catch (IllegalArgumentException e) {
				throw new ArgumentSyntaxException();
			}

			try {
				mBusSap.readMeter(deviceAddress, timeout);
			} catch (Exception e) {
				mBusSap.closeSerialPort();
				throw new ConnectionException(e);
			}

			connectionHandle = new ConnectionHandle(mBusSap);
			connections.put(interfaceAddress, connectionHandle);

		}
		else {

			try {
				connectionHandle.getMBusSap().readMeter(deviceAddress, timeout);
			} catch (Exception e) {
				throw new ConnectionException(e);
			}

			connectionHandle.increaseDeviceCounter();
		}

		return connectionHandle;
	}

	@Override
	public void disconnect(DeviceConnection connection) {
		ConnectionHandle connectionHandle = (ConnectionHandle) connection.getConnectionHandle();
		connectionHandle.decreaseDeviceCounter();
		if (connectionHandle.getDeviceCounter() == 0) {
			connectionHandle.getMBusSap().closeSerialPort();
			connections.remove(connection.getInterfaceAddress());
		}
	}

	@Override
	public Object read(DeviceConnection connection, List<ChannelRecordContainer> containers,
			Object containerListHandle, String samplingGroup, int timeout) throws UnsupportedOperationException,
			ConnectionException {

		MBusSap mBusSap = ((ConnectionHandle) connection.getConnectionHandle()).getMBusSap();

		VariableDataResponse response;
		try {
			response = mBusSap.readMeter(connection.getDeviceAddress(), timeout);
		} catch (IOException e1) {
			disconnect(connection);
			throw new ConnectionException(e1);
		} catch (TimeoutException e1) {
			for (ChannelRecordContainer container : containers) {
				container.setRecord(new Record(Flag.TIMEOUT));
			}
			return null;
		}

		long timestamp = System.currentTimeMillis();

		for (ChannelRecordContainer container : containers) {

			for (VariableDataBlock vdb : response.getVariableDataBlocks()) {

				String s = container.getChannelAddress();
				byte[] dib = vdb.getDIB();
				int len = dib.length * 2;
				for (int i = 0; i < len; i += 2) {
					try {
						if (dib[i / 2] != (byte) ((Character.digit(s.charAt(i), 16) << 4) + Character.digit(
								s.charAt(i + 1), 16))) {
							continue;
						}
					} catch (Exception e) {
						container.setRecord(new Record(Flag.DRIVER_ERROR_CHANNEL_ADDRESS_SYNTAX_INVALID));
						break;
					}
				}

				if (s.charAt(len) != ':') {
					container.setRecord(new Record(Flag.DRIVER_ERROR_CHANNEL_ADDRESS_SYNTAX_INVALID));
					break;
				}

				byte[] vib = vdb.getVIB();
				int len2 = vib.length * 2;
				for (int i = 0; i < len2; i += 2) {
					try {
						if (vib[i / 2] != (byte) ((Character.digit(s.charAt(len + 1 + i), 16) << 4) + Character.digit(
								s.charAt(len + i + 2), 16))) {
							continue;
						}
					} catch (Exception e) {
						container.setRecord(new Record(Flag.DRIVER_ERROR_CHANNEL_ADDRESS_SYNTAX_INVALID));
						break;
					}
				}

				try {
					vdb.parse();
				} catch (ParseException e) {
					container.setRecord(new Record(Flag.DRIVER_ERROR_CHANNEL_NOT_ACCESSIBLE));
					logger.debug("Unable to parse VariableDataBlock received via M-Bus", e);
					break;
				}

				container.setRecord(new Record(new DoubleValue(vdb.getScaledValue()), timestamp));
				break;
			}

			if (container.getRecord() == null) {
				container.setRecord(new Record(Flag.DRIVER_ERROR_CHANNEL_WITH_THIS_ADDRESS_NOT_FOUND));
			}

		}

		return null;
	}

	@Override
	public void startListening(DeviceConnection connection, List<ChannelRecordContainer> containers,
			RecordsReceivedListener listener) throws UnsupportedOperationException, ConnectionException {
		throw new UnsupportedOperationException(
				"OpenMUC Listening Mode not supported for MBus. Fix your configuration.");
	}

	@Override
	public Object write(DeviceConnection connection, List<ChannelValueContainer> containers,
			Object containerListHandle, int timeout) throws UnsupportedOperationException, ConnectionException {
		throw new UnsupportedOperationException("Unable to write Channel. MBus Driver is read only.");
	}
}
