/*
 * Copyright 2011-13 Fraunhofer ISE
 *
 * This file is part of OpenMUC.
 * For more information visit http://www.openmuc.org
 *
 * OpenMUC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * OpenMUC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with OpenMUC.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.framework.server.restws.channel;

import java.io.CharArrayWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.Writer;
import java.nio.ByteBuffer;
import java.util.List;

import org.openmuc.framework.data.ByteStringValue;
import org.openmuc.framework.data.DoubleValue;
import org.openmuc.framework.data.Record;
import org.openmuc.framework.data.TypeConversionException;
import org.openmuc.framework.data.Value;
import org.openmuc.framework.dataaccess.Channel;

import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonGenerationException;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.core.JsonParser;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.core.JsonToken;

/**
 * Helper class that transforms the objects communicated with the webservices in this package to and from JSON
 * 
 * @author Karsten Mueller-Bier
 */
public final class JsonWrapper {

	private final JsonFactory factory;

	public JsonWrapper() {
		factory = new JsonFactory();
	}

	/**
	 * Creates a JSON string from a single Record object
	 * 
	 * @see JsonWrapper#writeRecordToJson(Record, JsonGenerator)
	 * @param value
	 *            The record to transform
	 * @return JSON representation of the record
	 * @throws IOException
	 */
	public String fromRecord(Record value) throws IOException {
		Writer writer = new CharArrayWriter();
		JsonGenerator json = factory.createGenerator(writer).useDefaultPrettyPrinter();

		writeRecordToJson(value, json);

		json.close();
		return writer.toString();
	}

	/**
	 * Creates a JSON string containing all records given
	 * 
	 * @param values
	 *            The records to transform
	 * @return JSON representation of the records
	 * @throws IOException
	 */
	public String fromRecordList(List<Record> values) throws IOException {
		Writer writer = new CharArrayWriter();
		JsonGenerator json = factory.createGenerator(writer).useDefaultPrettyPrinter();

		json.writeStartArray();
		for (Record record : values) {
			writeRecordToJson(record, json);
		}
		json.writeEndArray();

		json.close();
		return writer.toString();
	}

	/**
	 * Creates a JSON string containing all channels given
	 * 
	 * @see JsonWrapper#writeRecordToJson(Record, JsonGenerator)
	 * @param channels
	 *            List of channels to write
	 * @return JSON representation of the channels
	 * @throws IOException
	 */
	public String fromChannelList(List<Channel> channels) throws IOException {
		Writer writer = new CharArrayWriter();
		JsonGenerator json = factory.createGenerator(writer).useDefaultPrettyPrinter();

		json.writeStartArray();
		for (Channel channel : channels) {
			json.writeStartObject();
			json.writeStringField("label", channel.getLabel());
			if (channel.getLatestSample() != null) {
				json.writeFieldName("record");
				writeRecordToJson(channel.getLatestSample(), json);
			}
			json.writeEndObject();
		}
		json.writeEndArray();
		json.close();

		return writer.toString();
	}

	/**
	 * Converts the value field of this JSON string to a org.openmuc.framework.data.Value object
	 * 
	 * @param textStream
	 *            Stream containing the JSON string
	 * @return The converted Value object
	 * @throws JsonProcessingException
	 *             If there is a problem extracting the value out of the JSON string
	 * @throws IOException
	 *             If there is a problem extracting the JSON String out of the InputStream
	 */
	public Value toValue(InputStream textStream) throws JsonProcessingException, IOException {
		JsonParser json = factory.createParser(textStream);

		Value result = null;
		while (json.nextToken() != JsonToken.START_OBJECT) {
		}
		while (json.nextToken() != JsonToken.END_OBJECT) {
			String fieldName = json.getCurrentName();
			json.nextToken();
			if ("value".equals(fieldName)) {
				if (json.getCurrentToken() == JsonToken.START_ARRAY) {
					ByteBuffer buffer = ByteBuffer.allocate(32);
					while (json.nextToken() != JsonToken.END_ARRAY) {
						buffer.put(json.getByteValue());
					}
					byte[] value = new byte[buffer.position()];
					buffer.flip();
					buffer.get(value);
					result = new ByteStringValue(value);
				}
				else {
					result = new DoubleValue(json.getDoubleValue());
				}
			}
		}

		return result;
	}

	/**
	 * Writes a Record object into a JSON Generator. The timestamp and flag are always written, the value is omitted of
	 * there is no value present
	 * 
	 * @param record
	 *            The Record object to transform
	 * @param json
	 *            The JSON Generator that shall create the JSON string
	 */
	private void writeRecordToJson(Record record, JsonGenerator json) throws JsonGenerationException, IOException {
		json.writeStartObject();
		json.writeObjectField("timestamp", record.getTimestamp());
		json.writeObjectField("flag", record.getFlag().name());
		if (record.getValue() != null) {
			try {
				json.writeNumberField("value", record.getValue().asDouble());
			} catch (TypeConversionException e) {
				json.writeArrayFieldStart("value");
				byte[] value = record.getValue().asByteArray();
				for (byte data : value) {
					json.writeNumber(data & 0xFF);
				}
				json.writeEndArray();
			}
		}
		json.writeEndObject();
	}
}
